package com.expd.apm.extension.agent.trace;

import java.util.Arrays;
import java.util.List;

import com.wily.introscope.agent.IAgent;
import com.wily.introscope.agent.stat.IDataAccumulator;
import com.wily.introscope.agent.stat.IIntegerFluctuatingCounterDataAccumulator;
import com.wily.introscope.agent.stat.ILongFluctuatingCounterDataAccumulator;
import com.wily.introscope.agent.stat.IStringEveryEventDataAccumulator;
import com.wily.introscope.agent.trace.ASingleMetricTracerFactory;
import com.wily.introscope.agent.trace.InvocationData;
import com.wily.introscope.agent.trace.ProbeIdentification;
import com.wily.introscope.stat.blame.BlameStackSnapshot;
import com.wily.util.properties.AttributeListing;

public class ParameterValueTracer extends ASingleMetricTracerFactory {
	private final static String kInteger = "int";
	private final static String kLong = "long";
	private final static String kString = "string";
	
	private final int fParamNum;
	private final String fParamType;
	private final IAgent fAgent; 
	private final List<?> fTypes = Arrays.asList(new String[]{ ParameterValueTracer.kInteger, ParameterValueTracer.kLong, ParameterValueTracer.kString });
	
	public ParameterValueTracer(IAgent agent, AttributeListing parameters, ProbeIdentification probe, Object sampleTracedObject) {
		super(agent, parameters, probe, sampleTracedObject);
		
		this.fAgent = agent;
		this.fParamNum = getIntegerParameter("parameterNumber", 0);
		this.fParamType = (getParameter("parameterType") == null) ? "int" : getParameter("parameterType");
		//getParameter("name")
		if(fTypes.contains(this.fParamType)) {
			this.fAgent.IAgent_getModuleFeedback().error("ParameterValueTracer: invalid parameter type; only int, long, and string are supported.");
		}
		
		getModuleFeedback().debug("ParameterValueTracer: initialized with following parameters: " 
				+ ", parameterNumber=" + this.fParamNum
				+ ", parameterType=" + this.fParamType
				+ ", useblame=" + this.calculateUseBlame(false));

	}

	protected final IDataAccumulator createDataAccumulator(String formattedName) {
		if(this.fParamType == null) return null;
		
		if(this.fParamType.equals(ParameterValueTracer.kInteger)) {
			return getDataAccumulatorFactory().safeGetIntegerFluctuatingCounterDataAccumulator(formattedName);
		} else if(this.fParamType.equals(ParameterValueTracer.kLong)) {
			return getDataAccumulatorFactory().safeGetLongFluctuatingCounterDataAccumulator(formattedName);		
		} else if(this.fParamType.equals(ParameterValueTracer.kString)) {
			return getDataAccumulatorFactory().safeGetStringEveryEventDataAccumulator(formattedName);
		} else {
			return null;
		} 
	}

	public final void ITracer_startTrace(int tracerIndex, InvocationData data) {
		if(!fTypes.contains(this.fParamType)) return;
		
		if (!(data.hasInvocationParametersAvailable())) {
			getModuleFeedback().error("ParameterValueTracer: Method parameters and return values are not available.....add \"WithParameters\" to directive");
		} else {
			BlameStackSnapshot snapshot = this.addToBlameStackIfEnabledAndReturnSnapshot(data);
			getModuleFeedback().debug("ParameterValueTracer: analyzing method parameters.");
			
			if(this.fParamType.equals(ParameterValueTracer.kInteger)) {
				recordInteger(data, snapshot);
			} else if(this.fParamType.equals(ParameterValueTracer.kLong)) {
				recordLong(data, snapshot);
			} else if(this.fParamType.equals(ParameterValueTracer.kString)) {
				recordString(data, snapshot);
			} 			
		}
	}
	
	private void recordInteger(InvocationData data, BlameStackSnapshot snapshot) {
		final IIntegerFluctuatingCounterDataAccumulator counter = (IIntegerFluctuatingCounterDataAccumulator)createDataAccumulator(formatParameterizedName(data));
		if(counter == null || counter.IDataAccumulator_isShutOff()) {
			return;
		}
		Integer param = (Integer)data.getInvocationParameterAsObject(fParamNum);
		
		if(param != null) {
			if(snapshot != null) {
				counter.IIntegerCounterDataAccumulator_setValue(param, snapshot);
			} else {
				counter.IIntegerCounterDataAccumulator_setValue(param);
			}
		} 
	}

	private void recordLong(InvocationData data, BlameStackSnapshot snapshot) {
		final ILongFluctuatingCounterDataAccumulator counter = (ILongFluctuatingCounterDataAccumulator)createDataAccumulator(formatParameterizedName(data));
		if(counter == null || counter.IDataAccumulator_isShutOff()) {
			return;
		}
		Long param = (Long)data.getInvocationParameterAsObject(fParamNum);
		
		if(param != null) {
			if(snapshot != null) {
				counter.ILongCounterDataAccumulator_setValue(param, snapshot);
			} else {
				counter.ILongCounterDataAccumulator_setValue(param);
			}
		} 		
	}

	private void recordString(InvocationData data, BlameStackSnapshot snapshot) {
		final IStringEveryEventDataAccumulator counter = (IStringEveryEventDataAccumulator)createDataAccumulator(formatParameterizedName(data));
		if(counter == null || counter.IDataAccumulator_isShutOff()) {
			return;
		}
		String param = data.getInvocationParameterAsObject(fParamNum).toString();
		
		if(param != null) {
			counter.IStringEveryEventDataAccumulator_addString(param);
		} 		
	}

	public final void ITracer_finishTrace(int tracerIndex, InvocationData data){
		if(!fTypes.contains(this.fParamType)) return;
		
		removeFromBlameStackIfEnabled(data);
	}
}